package cn.com.qmth.examcloud.core.oe.student.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.core.oe.student.base.utils.Check;
import cn.com.qmth.examcloud.core.oe.student.base.utils.CommonUtil;
import cn.com.qmth.examcloud.core.oe.student.bean.client.FaceLiveVerifyInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.client.FaceLiveVerifyResult;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamFaceLiveVerifyRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamFaceLiveVerifyEntity;
import cn.com.qmth.examcloud.core.oe.student.dao.enums.FaceLiveVerifyStatus;
import cn.com.qmth.examcloud.core.oe.student.service.ExamFaceLiveVerifyService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamingSessionService;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.cache.bean.ExamPropertyCacheBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.IsSuccess;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.examing.ExamingHeartbeat;
import cn.com.qmth.examcloud.support.examing.ExamingSession;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.util.Date;
import java.util.List;
import java.util.Optional;

@Service
public class ExamFaceLiveVerifyServiceImpl implements ExamFaceLiveVerifyService {

    private static final Logger log = LoggerFactory.getLogger(ExamFaceLiveVerifyServiceImpl.class);

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamingSessionService examingSessionService;

    @Autowired
    private ExamFaceLiveVerifyRepo examFaceLiveVerifyRepo;

    @Autowired
    private RedisClient redisClient;

    @Override
    public FaceLiveVerifyInfo startFaceLiveVerify(Long examRecordDataId, Long studentId) {
        Check.isNull(studentId, "学生ID不能为空");
        Check.isNull(examRecordDataId, "考试记录ID不能为空");

        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordData == null || ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            throw new StatusException("考试记录无效");
        }

        ExamingSession examingSession = examingSessionService.getExamingSession(studentId);
        if (examingSession == null) {
            throw new StatusException("考试会话已过期");
        }

        int times = 0;
        boolean hasSuccess = false;
        Long faceLiveVerifyId = null;

        List<ExamFaceLiveVerifyEntity> entities = examFaceLiveVerifyRepo.findByExamRecordDataId(examRecordData.getId());
        if (!CollectionUtils.isEmpty(entities)) {
            for (ExamFaceLiveVerifyEntity entity : entities) {
                if (FaceLiveVerifyStatus.SUCCESS == entity.getStatus()) {
                    hasSuccess = true;
                }

                if (entity.getFinished()) {
                    times++;
                } else {
                    faceLiveVerifyId = entity.getId();
                }
            }
        }

        FaceLiveVerifyInfo info = new FaceLiveVerifyInfo();
        info.setTimes(times);
        info.setHasSuccess(hasSuccess);

        if (times > 1) {
            log.warn("活检次数已超过2次！examRecordDataId = {}, studentId = {}", examRecordDataId, studentId);
            return info;
        }

        if (hasSuccess) {
            log.warn("活检已检测成功过！examRecordDataId = {}, studentId = {}", examRecordDataId, studentId);
            return info;
        }

        // 如果已经人脸检测过一次且未成功,再安排一次检测
        if (faceLiveVerifyId == null) {
            ExamFaceLiveVerifyEntity entity = new ExamFaceLiveVerifyEntity();
            entity.setExamRecordDataId(examRecordDataId);
            entity.setFinished(false);
            entity.setStatus(FaceLiveVerifyStatus.ERROR);
            examFaceLiveVerifyRepo.save(entity);
            faceLiveVerifyId = entity.getId();
            times++;
        }
        info.setFaceLiveVerifyId(faceLiveVerifyId);
        info.setTimes(times);

        Integer startMinute = this.calculateStartFaceVerifyMinute(examRecordDataId);
        info.setStartMinute(startMinute);
        return info;
    }

    @Override
    public void saveFaceLiveVerifyResult(FaceLiveVerifyResult req) {
        Check.isNull(req.getStudentId(), "学生ID不能为空");
        Check.isNull(req.getFaceLiveVerifyId(), "当前活检记录ID不能为空");
        Check.isNull(req.getExamRecordDataId(), "考试记录ID不能为空");
        Check.isNull(req.getStatus(), "人脸活体验证状态不能为空");
        Check.isEmpty(req.getActions(), "动作验证列表不能为空");

        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(req.getExamRecordDataId());
        if (examRecordData == null || ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            log.warn("考试记录无效！examRecordDataId = {}, studentId = {}", req.getExamRecordDataId(), req.getStudentId());
            return;
        }

        ExamingSession examingSession = examingSessionService.getExamingSession(req.getStudentId());
        if (examingSession == null) {
            log.warn("考试会话已过期！examRecordDataId = {}, studentId = {}", req.getExamRecordDataId(), req.getStudentId());
            return;
        }

        Optional<ExamFaceLiveVerifyEntity> optional = examFaceLiveVerifyRepo.findById(req.getFaceLiveVerifyId());
        if (!optional.isPresent()) {
            log.warn("当前活检记录不存在！faceLiveVerifyId = {}", req.getFaceLiveVerifyId());
            return;
        }

        ExamFaceLiveVerifyEntity entity = optional.get();
        if (entity.getFinished()) {
            log.warn("当前活检记录已完成！faceLiveVerifyId = {}", req.getFaceLiveVerifyId());
            return;
        }

        entity.setFinished(true);
        entity.setStatus(req.getStatus());
        entity.setFaceCount(req.getFaceCount());
        entity.setSimilarity(req.getSimilarity());
        entity.setRealness(req.getRealness());
        entity.setErrorMsg(req.getErrorMsg());
        entity.setProcessTime(req.getProcessTime() != null ? req.getProcessTime() : 1L);
        entity.setActions(new JsonMapper().toJson(req.getActions()));
        entity.setUpdateTime(new Date());
        examFaceLiveVerifyRepo.save(entity);

        examRecordData.setFaceVerifyResult(FaceLiveVerifyStatus.SUCCESS == req.getStatus() ? IsSuccess.SUCCESS : IsSuccess.FAILED);
        if (FaceLiveVerifyStatus.SUCCESS != req.getStatus()) {
            // 活检失败，则认为违纪
            examRecordData.setIsIllegality(true);
        }
        examRecordDataService.saveExamRecordDataCache(req.getExamRecordDataId(), examRecordData);
    }

    @Override
    public Integer calculateStartFaceVerifyMinute(Long examRecordDataId) {
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordData == null || ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            throw new StatusException("考试记录无效");
        }

        //开启了人脸检测
        if (!FaceBiopsyHelper.isFaceVerify(examRecordData.getRootOrgId(), examRecordData.getExamId(), examRecordData.getStudentId())) {
            //如果未开启人脸活体检测，则返回null
            return null;
        }

        ExamingSession examingSession = examingSessionService.getExamingSession(examRecordData.getStudentId());
        if (examingSession == null) {
            throw new StatusException("考试会话已过期");
        }

        // 活体检测开始分钟数
        int faceVerifyStartMinute = this.getExamPropertyValue(examRecordData.getExamId(),
                examRecordData.getStudentId(), ExamProperties.FACE_VERIFY_START_MINUTE);

        // 活体检测结束分钟数（默认 <= 交卷冻结时间）
        int faceVerifyEndMinute = this.getExamPropertyValue(examRecordData.getExamId(),
                examRecordData.getStudentId(), ExamProperties.FACE_VERIFY_END_MINUTE);

        String examingHeartbeatKey = RedisKeyHelper.getBuilder().examingHeartbeatKey(examRecordDataId);
        ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey, ExamingHeartbeat.class);

        // 考试已用时间
        int usedMinute = 0;
        if (examingHeartbeat != null) {
            usedMinute = Math.toIntExact(examingHeartbeat.getCost() / 60L);
        }

        if (usedMinute < faceVerifyStartMinute) {
            // 考试已用时间 尚未达到 配置的活体检测开始分钟数时，结果 = Random(开始分钟数-结束分钟数)-考试已用时间
            return CommonUtil.calculationRandomNumber(faceVerifyStartMinute, faceVerifyEndMinute) - usedMinute;
        } else if (usedMinute >= faceVerifyStartMinute && usedMinute < faceVerifyEndMinute) {
            // 考试已用时间 介于 配置的活体检测开始和结束分钟数中间时，结果 = Random(考试已用时间-结束分钟数)-考试已用时间，默认最小1分钟
            int faceVerifyMinute = CommonUtil.calculationRandomNumber(usedMinute, faceVerifyEndMinute) - usedMinute;
            return faceVerifyMinute > 1 ? faceVerifyMinute : 1;
        } else if (usedMinute >= faceVerifyEndMinute) {
            // 考试已用时间 大于 配置的活体检测结束分钟数时，结果 = Random(1,4)
            return CommonUtil.calculationRandomNumber(1, Constants.MAX_FACE_LIVE_VERIFY_MINUTE);
        }

        return null;
    }

    private int getExamPropertyValue(Long examId, Long studentId, ExamProperties propKey) {
        ExamPropertyCacheBean property = ExamCacheTransferHelper.getCachedExamProperty(examId, studentId, propKey.name());
        if (property == null || StringUtils.isEmpty(property.getValue())) {
            throw new StatusException(propKey.getDesc() + "未设置");
        }
        return Integer.parseInt(property.getValue());
    }

}
